// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Developer             Changes
//
// 16 Jan 2003   Tom Barry		       Use AvisynthPluginInit2

// 02 Nov 2002   Tom Barry		       Create UnDot V 0.0.1.0
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////
#include "internal.h"
#include "UnDot.h"


class UnDot : public GenericVideoFilter {
    int  Max_Frames;
	bool SSEMMXenabled;

// Only for YUY2
    BYTE* dstp;
    int src_pitch;
    int dst_pitch;
    int row_size;
    int RowPixels;
    int height;
    int FldHeight;

// Only for YV12



	BOOL UnDotYUY2(int src_pit, int dst_pit,
			int row_size, const BYTE* srcp, BYTE* dstp, int FldHeight);

	BOOL UnDotYV12(int src_pit, int dst_pit,
			int row_size, const BYTE* srcp, BYTE* dstp, int FldHeight);

public:
    UnDot(PClip _child, IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
	    Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;

	    if (!vi.IsYUY2() && !vi.IsYV12())
		{
			env->ThrowError("UnDot: Supports YUY2 color format only");
		}

		if (!SSEMMXenabled)
		{
			env->ThrowError("UnDot: Only SSEMMX CPU's currently supported");
		}
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall UnDot::GetFrame(int inFrame, IScriptEnvironment* env) 
{
	PVideoFrame src = child->GetFrame(inFrame, env);
	PVideoFrame dst = env->NewVideoFrame(vi);

// Only for YUY2
    BYTE* dstp;
	const BYTE* srcp;
    int src_pitch;
    int dst_pitch;
    int row_size;
    int RowPixels;
    int height;

// Only for YV12
	const BYTE* srcpY;
	const BYTE* srcpV;   
	const BYTE* srcpU;

	BYTE* dstpY;
	BYTE* dstpV;   
	BYTE* dstpU;

	int src_pitchY;
	int src_pitchUV;
	int dst_pitchY;
	int dst_pitchUV;

	int row_sizeY;   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	int row_sizeUV;  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   

	int heightY;
	int heightUV;
//

	if (vi.IsYUY2())
	{
		srcp = src->GetReadPtr();

		if (inFrame < 1 || inFrame >= Max_Frames-1)
		{
			return src;	
		}
		dst = env->NewVideoFrame(vi);
		dstp = dst->GetWritePtr();
		dstp = dst->GetWritePtr();
		src_pitch = src->GetPitch();
		dst_pitch = dst->GetPitch();
		row_size = dst->GetRowSize();
		RowPixels = row_size >> 1;
		height = dst->GetHeight();

		UnDotYUY2(src_pitch, dst_pitch,
			row_size, srcp, dstp, height);					
		
	}

	else										// We already eliminated RGB above, so YV12
	{
		srcpY = src->GetReadPtr(PLANAR_Y);
		srcpV = src->GetReadPtr(PLANAR_V);   
		srcpU = src->GetReadPtr(PLANAR_U);

		dstpY = dst->GetWritePtr(PLANAR_Y);
		dstpV = dst->GetWritePtr(PLANAR_V);   
		dstpU = dst->GetWritePtr(PLANAR_U);

		src_pitchY = src->GetPitch(PLANAR_Y);
		src_pitchUV = src->GetPitch(PLANAR_V);
		dst_pitchY = dst->GetPitch(PLANAR_Y);
		dst_pitchUV = dst->GetPitch(PLANAR_U);

		row_sizeY = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
		row_sizeUV = dst->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   

		heightY = dst->GetHeight(PLANAR_Y);
		heightUV = dst->GetHeight(PLANAR_U);

		UnDotYV12(src_pitchY, dst_pitchY,
			row_sizeY, srcpY, dstpY, heightY);					

		UnDotYV12(src_pitchUV, dst_pitchUV,
			row_sizeUV, srcpU, dstpU, heightUV);					

		UnDotYV12(src_pitchUV, dst_pitchUV,
			row_sizeUV, srcpV, dstpV, heightUV);					
	}

	__asm emms;
	return dst;
}


// YUY2 has 2 bytes per pixel
#define BPP 2	
BOOL	UnDot::UnDotYUY2(int src_pit, int dst_pit,
			int row_size, const BYTE* srcp, BYTE* dstp,int FldHeight)	
{

const BYTE* pSrc = srcp;
BYTE* pDest = dstp;
int y;
		
	memcpy(pDest, pSrc, row_size);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), row_size);  // and last
	pDest += dst_pit;

	for (y=1; y <= FldHeight-2; y++)	
	{
		__asm
		{
		// Loop general reg usage
		//
		// ecx - count
		// edi - dest pixels
		// esi - src pixels, 1 line up

		mov		esi, pSrc
		mov		ebx, src_pit 
		mov		edi, pDest 
		mov		eax, row_size
		mov		ecx, eax
		shr		ecx, 3						// do 8 bytes at a time

		// do first  qword
		movq	mm0, qword ptr[esi+ebx]		// move 1st 4 pixels
		movq	qword ptr[edi], mm0
		add		esi, BPP					// skip over first pixel
		GetMinMax	
		pand	mm0, YMask					// allow chroma min 0
		por		mm1, UVMask					// allow chroma max 255
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movq	qword ptr[edi+BPP], mm0

		// do last qword
		lea		esi, [esi+eax-8-BPP]		// point at last qword
		movq	mm0, qword ptr[esi+ebx]		// move 1st 4 pixels
		movq	qword ptr[edi+eax-8], mm0
		sub		esi, BPP					// back up one pixel
		GetMinMax	
		pand	mm0, YMask					// allow chroma min 0
		por		mm1, UVMask					// allow chroma max 255
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movq	qword ptr[edi+eax-BPP-8], mm0

		mov		esi, pSrc					// restore esi
		sub		ecx, 1						// but both ends done manually
		jz		AllDoneYUY2					// too short, don't bother
		

LoopYUY2:	
		add		esi, 8						// bump to next qword
		add		edi, 8

		GetMinMax
		pand	mm0, YMask					// allow chroma min 0
		por		mm1, UVMask					// allow chroma max 255
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movntq	qword ptr[edi], mm0			// and store it
		loop	LoopYUY2

AllDoneYUY2:	
		}

	// adjust for next line
	pSrc  += src_pit;
	pDest += dst_pit;
	}
	return 0;
}


// YV12 has 1 bytes per pixel
#undef	BPP
#define BPP 1	
BOOL	UnDot::UnDotYV12(int src_pit, int dst_pit,
			int row_size, const BYTE* srcp, BYTE* dstp,int FldHeight)	
{

const BYTE* pSrc = srcp;
BYTE* pDest = dstp;
int y;
		
	memcpy(pDest, pSrc, row_size);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), row_size);  // and last
	pDest += dst_pit;

	for (y=1; y <= FldHeight-2; y++)	
	{
		__asm
		{
		// Loop general reg usage
		//
		// ecx - count
		// edi - dest pixels
		// esi - src pixels, 1 line up

		mov		esi, pSrc
		mov		ebx, src_pit 
		mov		edi, pDest 
		mov		eax, row_size
		mov		ecx, eax
		shr		ecx, 3						// do 8 bytes at a time

		// do first  qword
		movq	mm0, qword ptr[esi+ebx]		// move 1st 4 pixels
		movq	qword ptr[edi], mm0
		add		esi, BPP					// skip over first pixel
		GetMinMax	
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movq	qword ptr[edi+BPP], mm0

		// do last qword
		lea		esi, [esi+eax-8-BPP]		// point at last qword
		movq	mm0, qword ptr[esi+ebx]		// move 1st 4 pixels
		movq	qword ptr[edi+eax-8], mm0
		sub		esi, BPP					// back up one pixel
		GetMinMax	
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movq	qword ptr[edi+eax-BPP-8], mm0

		mov		esi, pSrc					// restore esi
		sub		ecx, 1						// but both ends done manually
		jz		AllDoneYV12					// too short, don't bother
		
LoopYV12:	
		add		esi, 8						// bump to next qword
		add		edi, 8

		GetMinMax
		pmaxub  mm0, qword ptr[esi+ebx]		// allow no lower than min
		pminub  mm0, mm1					// allow no higher than max
		movntq	qword ptr[edi], mm0			// and store it
		loop	LoopYV12

AllDoneYV12:	
		}

	// adjust for next line
	pSrc  += src_pit;
	pDest += dst_pit;
	}
	return 0;
}


AVSValue __cdecl Create_UnDot(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new UnDot(args[0].AsClip(), env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("UnDot", "c", Create_UnDot, 0);
    return "`UnDot' Remove stray dots, dering";
}

